let fl = Str.regexp "p cnf \\([0-9]+\\) \\([0-9]+\\)"
let var = Str.regexp "\\(-?[0-9]+\\)"


let parse_fl s =
  if Str.string_match fl s 0 then
    let nvar = int_of_string (Str.matched_group 1 s) in
    nvar
  else
    failwith ("Bad first line "^s)


let parse_line s =

  let rec pl acc i =
    let i' = Str.search_forward var s i in
    let all = Str.matched_group 0 s in
    let v = int_of_string (Str.matched_group 1 s) in
    if v = 0 then
      acc
    else
      pl (v::acc) (i'+String.length all) in

  pl [] 0


let abs n = if n > 0 then n else -n


let print_var out n = out ("c"^(string_of_int (abs n)))


let print_lit out n =
  if n > 0 then
    print_var out n
  else (
    out "(1-"; print_var out n; out ")"
  )


let print_beginning out =
  out "#include <p8q.h>\n\nint\nmain(void)\n{\n"


let print_fl out nvar =
  for i = 1 to nvar do
    out "\n  int ";
    print_var out i;
    out " = read_private_int(1);"
  done;
  out "\n"


let print_line out l =
  out "\n  assert (";
  match l with
    | [] -> assert false
    | x::xs -> (
      print_lit out x;
      List.iter (fun n -> out " + "; print_lit out n) xs
    );
  out ");"


let print_end out =
  out "\n  return 0;\n}"


let translator input_name =
  let output_name = (Filename.chop_extension input_name)^".c" in
  let input = open_in input_name in
  let output = open_out output_name in
  let outref = ref 0 in
  let out s = output_string output s; incr outref; if !outref > 999 then (outref := 0; flush output) in

  print_beginning out;
  let fl = ref true in
  try
    while true do
      let s = input_line input in
      if (String.length s <= 0 || s.[0] <> 'c') then
        if !fl then (
          fl := false;
          let nvar = parse_fl s in
          print_fl out nvar
        ) else if s.[0] <> 'c' then (
          let l = parse_line s in
          print_line out l
        )
    done;
  with
    | End_of_file ->
      print_end out;
      close_in input;
      close_out output


let _ = translator Sys.argv.(1)
