with import <nixpkgs> {};

let ocamlPackages = ocaml-ng.ocamlPackages_4_02; in

let coq = callPackage ./nix/coq_8_4.nix {
  inherit (ocamlPackages) ocaml findlib camlp5;
}; in

let callPackage = newScope { inherit coq; }; in

let compcert = callPackage ./nix/compcert.nix {
  tools = pkgsi686Linux.gcc;
  inherit (ocamlPackages) menhir;
}; in

let ppsimpl = callPackage ./nix/ppsimpl.nix {}; in

let mathcomp = callPackage ./nix/mathcomp.nix {}; in

let coqprime = callPackage ./nix/coqprime.nix {}; in

stdenv.mkDerivation {
  name = "p8q";
  CompCert = "${compcert}/";
  buildInputs = [ coq ppsimpl coqprime compcert mathcomp ]
    ++ (with ocamlPackages; [ ocaml findlib menhir ]);

  src = ./.;

  configurePhase = ''
    ln -s "${compcert}/lib/coq/${coq.coq-version}/user-contrib/compcert/" cc
    coq_makefile -o Makefile -f _CoqProject
  '';

  buildTargets = [ "rtlc.native" "qapRead" ];

  installPhase = ''
    make COQLIB=$out/lib/coq/${coq.coq-version}/ install
    mkdir -p $out/bin
    mv _build/frontend/rtlc.native $out/bin/rtlc
    mv _build/frontend/QapRead.native $out/bin/qapRead
  '';
}
