Require Import Utf8 BinPos.
Require Import RField.
Import EquivDec.

Require Import ssreflect ssrnat ssrbool eqtype fintype ssrfun seq.
Require Import div bigop prime ssralg zmodp.
Import GRing.Theory.

Unset Elimination Schemes.
Set Implicit Arguments.


Module FieldSpec.

  Section S.

    Context f feq `(EqDec f feq) `(Plus f) `(Zero f) `(Opposite f) `(Mult f) `(One f) `(Inverse f).

    Variable F : fieldType.
    Variable embed : f → F.

    Record spec : Prop :=
      { embed_eq : ∀ x y, (x === y) <-> (embed x = embed y)
        ; embed_zero : embed zero = 0%R
        ; embed_one : embed one = 1%R
        ; embed_plus : ∀ x y, embed (plus x y) = ((embed x) + (embed y))%R
        ; embed_opp : ∀ x, embed (opposite x) = (- (embed x))%R
        ; embed_mult : ∀ x y, embed (mult x y) = ((embed x) * (embed y))%R
        ; embed_inv : ∀ x, embed x != 0%R → embed (RField.inverse x) = ((embed x)^-1)%R
      }.

  End S.

  Arguments spec {_ _ _ _ _ _ _ _ _} _ _.

  Ltac embed_tac_field Hf :=
    repeat (
        rewrite (embed_eq Hf) ||
        rewrite (embed_zero Hf) ||
        rewrite (embed_one Hf) ||
        rewrite (embed_plus Hf) ||
        rewrite (embed_opp Hf) ||
        rewrite (embed_mult Hf) ||
        rewrite (embed_inv Hf)
      ).

End FieldSpec.


Require Import LinearCombination.

Module Type LCSpec.
  Include LinearCombinationType.

  Import FieldSpec.

  Section LCSpec.

    Context f feq `(EqDec f feq) `(Plus f) `(Zero f) `(Opposite f) `(Mult f) `(One f) `(Inverse f).

    Variable F : fieldType.
    Variable embed : f → F.
    Variable embed_spec : spec F embed.

    Parameter coeff : t f → var → f.
    Axiom denote_ad : ∀ ω ρ (a:t f),
        embed (den ω ρ a) = (\sum_(1 <= k < (Pos.to_nat ω).+1) ((embed (ρ (Pos.of_nat k))) * (embed (coeff a (Pos.of_nat k)))))%R.

  End LCSpec.
End LCSpec.


Require LCList QAP.

Module LCListSpec <: LinearCombinationType <: LCSpec.
  Include QAP.LC.

  Import FieldSpec.

  Section LCSpec.

    Context f feq `(EqDec f feq) `(Plus f) `(Zero f) `(Opposite f) `(Mult f) `(One f) `(Inverse f).

    Variable F : fieldType.
    Variable embed : f → F.
    Variable embed_spec : spec F embed.

    (* Definition coeff (a:t f) (x:var) : f := *)
    (*   List.fold_left (fun acc (kv:(var * f)%type) => let (k,v) := kv in if Pos.eqb k x then v else acc) a 0. *)
    (* TODO: fill this *)
    Parameter coeff : t f → var → f.
    Axiom denote_ad : ∀ ω ρ (a:t f),
        embed (den ω ρ a) = (\sum_(1 <= k < (Pos.to_nat ω).+1) ((embed (ρ (Pos.of_nat k))) * (embed (coeff a (Pos.of_nat k)))))%R.

  End LCSpec.
End LCListSpec.


(* Not used (yet) *)
(* Require LCMap. *)

(* Module LCMapSpec : LCSpec. *)
(*   Include LCMap.LC. *)

(*   Import FieldSpec. *)

(*   Section LCSpec. *)

(*     Context f feq `(EqDec f feq) `(Plus f) `(Zero f) `(Opposite f) `(Mult f) `(One f) `(Inverse f). *)

(*     Variable F : fieldType. *)
(*     Variable embed : f → F. *)
(*     Variable embed_spec : spec F embed. *)

(*     (* TODO: fill this *) *)
(*     Parameter coeff : t f → var → f. *)
(*     Axiom denote_ad : ∀ ω ρ (a:t f), *)
(*         embed (den ω ρ a) = (\sum_(1 <= k < (Pos.to_nat ω).+1) ((embed (ρ (Pos.of_nat k))) * (embed (coeff a (Pos.of_nat k)))))%R. *)

(*   End LCSpec. *)
(* End LCMapSpec. *)







(*     Section S. *)

(*       Variable var : eqType. *)

(*       Lemma uniq_rem x (f: seq var) : *)
(*         uniq f → *)
(*         x \notin (rem x f). *)
(*       Proof. *)
(*         elim: f => [|y f IHf] //=. *)
(*         - move/andP => [[H1] H2]. case e₁: (y == x). *)
(*           * by move/eqP: e₁ => <-. *)
(*           * rewrite in_cons (eq_sym x y) e₁ /=. by apply IHf. *)
(*       Qed. *)

(*       Definition denotep (ρ : var → F) (p : var → F) (f : seq var) : F := *)
(*         (\sum_(v <- f) ((p v) * (ρ v)))%R. *)

(*       Lemma denotep_nil ρ p : denotep ρ p nil = 0%R. *)
(*       Proof. by rewrite/denotep big_nil. Qed. *)

(*       Lemma denotep_cons ρ p x xs : *)
(*         denotep ρ p (x :: xs) = (((p x) * (ρ x)) + (denotep ρ p xs))%R. *)
(*       Proof. by rewrite/denotep big_cons. Qed. *)

(*       Lemma denotep_cat ρ p f₁ f₂: *)
(*         denotep ρ p (f₁ ++ f₂) = ((denotep ρ p f₁) + (denotep ρ p f₂))%R. *)
(*       Proof. by rewrite/denotep big_cat. Qed. *)

(*       Definition restr (p : var → F) (f : seq var) : var → F := *)
(*         fun x => if x \in f then p x else 0%R. *)

(*       Lemma restr_restr p f x: *)
(*         restr (restr p f) f x = restr p f x. *)
(*       Proof. by rewrite/restr; case: (x \in f). Qed. *)

(*       Lemma denotep_restr ρ p f : *)
(*         denotep ρ p f = denotep ρ (restr p f) f. *)
(*       Proof. *)
(*         rewrite/denotep 2!big_seq. apply eq_bigr => x Hx. by rewrite/restr Hx. *)
(*       Qed. *)

(*       Lemma restr_nil ρ p f : denotep ρ (restr p nil) f = 0%R. *)
(*       Proof. *)
(*         rewrite/denotep. apply big_rec => //=. *)
(*         by rewrite/restr => x _ _ -> /=; rewrite mul0r add0r. *)
(*       Qed. *)

(*       Lemma denotep_restr_notin ρ p f₁ f₂ *)
(*         (H: forall x, (x \in f₂) → (x \in f₁ = false)) : *)
(*         denotep ρ (restr p f₁) f₂ = 0%R. *)
(*       Proof. *)
(*         rewrite/denotep big_seq. apply big_rec => //= x _ H1 ->. *)
(*         move: (H _ H1). rewrite /restr => ->. by rewrite mul0r add0r. *)
(*       Qed. *)

(*       Fixpoint union (f₁ f₂ : list var) := *)
(*         match f₁ with *)
(*           | nil => f₂ *)
(*           | x::xs => if x \in f₂ then union xs f₂ else union xs (x::f₂) *)
(*         end. *)

(*       Lemma union_uniq f₁ : ∀ f₂, uniq f₂ -> uniq (union f₁ f₂). *)
(*       Proof. *)
(*         elim: f₁ => [|x xs IHxs] /= f₂ Hf₂ //=. *)
(*         - case e₁: (x \in f₂). *)
(*           * by apply IHxs. *)
(*           * by apply IHxs; rewrite /= e₁. *)
(*       Qed. *)

(*       Lemma union_in x f₁ : ∀ f₂, *)
(*         x \in (union f₁ f₂) = (x \in f₁) || (x \in f₂). *)
(*       Proof. *)
(*         elim: f₁ => [|y f₁ IHf₁] //= f₂. *)
(*         - rewrite in_cons. case e₁: (y \in f₂). *)
(*           * rewrite IHf₁. case e₂: (x == y) => //=. *)
(*             + move/eqP: e₂ => ->. by rewrite e₁ orb_true_r. *)
(*           * by rewrite IHf₁ in_cons orb_comm -orb_assoc (orb_comm (x \in f₂) (x \in f₁)) orb_assoc. *)
(*       Qed. *)

(*       Lemma restr_restr_union₁ x p f₁ f₂ : *)
(*         restr (restr p f₁) (union f₁ f₂) x = restr p f₁ x. *)
(*       Proof. *)
(*         rewrite/restr union_in. case: (x \in f₁) => //=. by case: (x \in f₂). *)
(*       Qed. *)

(*       Lemma restr_restr_union₂ x p f₁ f₂: *)
(*         restr (restr p f₂) (union f₁ f₂) x = restr p f₂ x. *)
(*       Proof. *)
(*         rewrite/restr union_in. case: (x \in f₂) => /=. *)
(*         - by rewrite orb_true_r. *)
(*         - by case: (x \in f₁). *)
(*       Qed. *)

(*       Lemma denotep_union₂_gen ρ p f₁: ∀ f₂ f₃, *)
(*         (forall x, (x \in f₃) → (x \in f₂ = false)) → *)
(*         denotep ρ p f₂ = denotep ρ (restr p f₂) (union f₁ (f₃ ++ f₂)). *)
(*       Proof. *)
(*         elim: f₁ => [|x xs IHxs] /= f₂ f₃ H. *)
(*         - rewrite denotep_cat. rewrite (denotep_restr_notin _ _ _ _ H) add0r. by apply denotep_restr. *)
(*         - rewrite mem_cat. case e₁: (x \in f₃). *)
(*           * rewrite (H _ e₁) /=. by apply IHxs. *)
(*           * case e₂: (x \in f₂) => /=. *)
(*             + by apply IHxs. *)
(*             + apply (IHxs _ (x::f₃)) => y. rewrite in_cons. move/orP => [/eqP -> //=|H1]. by apply H. *)
(*       Qed. *)

(*       Lemma denotep_union₂ ρ p f₁ f₂ : *)
(*         denotep ρ p f₂ = denotep ρ (restr p f₂) (union f₁ f₂). *)
(*       Proof. by apply (denotep_union₂_gen _ _ _ _ nil). Qed. *)

(*       Lemma denotep_restr_cons ρ p x f₁ f₂ *)
(*         (Hx: x \in f₂ = false) : *)
(*         denotep ρ (restr p f₁) f₂ = denotep ρ (restr p (x :: f₁)) f₂. *)
(*       Proof. *)
(*         rewrite/denotep 2!big_seq. apply eq_bigr => y Hy. rewrite/restr in_cons. case: (y \in f₁). *)
(*         - by rewrite orb_true_r. *)
(*         - case e₁: (y == x) => //=. *)
(*           by move/eqP: e₁ Hy Hx => -> ->. *)
(*       Qed. *)

(*       Lemma denotep_restr_cons2 ρ p x f₁ f₂ *)
(*             (H1 : x \notin f₁) *)
(*             (H2 : x \in f₂) *)
(*             (H3 : uniq f₂) : *)
(*         (p x * ρ x + denotep ρ (restr p f₁) f₂)%R = *)
(*         denotep ρ (restr p (x :: f₁)) f₂. *)
(*       Proof. *)
(*         rewrite/denotep 2!(big_rem _ H2) /=. *)
(*         replace (restr p f₁ x) with (0%R : F). *)
(*         - replace (restr p (x :: f₁) x) with (p x). *)
(*           * rewrite mul0r add0r. move: (denotep_restr_cons ρ p x f₁ (rem x f₂)). rewrite/denotep => -> //=. *)
(*             by apply negbTE, uniq_rem. *)
(*           * by rewrite/restr in_cons eqxx. *)
(*         - by rewrite/restr (negbTE H1). *)
(*       Qed. *)

(*       Lemma denotep_union₁ f₁: ∀ ρ p f₂, *)
(*         uniq f₁ → uniq f₂ → *)
(*         denotep ρ p f₁ = denotep ρ (restr p f₁) (union f₁ f₂). *)
(*       Proof. *)
(*         elim: f₁ => [|x f₁ IHf₁] /= ρ p f₂ //=. *)
(*         - by rewrite denotep_nil restr_nil. *)
(*         - move/andP => [H1 H2] H3. rewrite denotep_cons. case e₁: (x \in f₂). *)
(*           * rewrite (IHf₁ _ _ _ H2 H3). apply denotep_restr_cons2 => //=. *)
(*             + by rewrite union_in e₁ orb_true_r. *)
(*             + by apply union_uniq. *)
(*           * assert (H4: uniq (x::f₂)). *)
(*               by rewrite /= (negbT e₁). *)
(*             rewrite (IHf₁ _ _ _ H2 H4). *)
(*             apply denotep_restr_cons2 => //=. *)
(*             + by rewrite union_in in_cons eqxx orb_true_r. *)
(*             +  by apply union_uniq. *)
(*       Qed. *)

(*     End S. *)

(*     Arguments denotep {var} _ _ _. *)
(*     Arguments restr {var} _ _ _. *)
(*     Arguments union {var} _ _. *)
(*     Arguments denotep_union₁ {var} _ _ _ _ _ _. *)
(*     Arguments denotep_union₂ {var} _ _ _ _. *)
(*     Arguments union_uniq {var} _ _ _. *)

(*     Context f feq `(EqDec f feq) `(Plus f) `(Zero f) `(Opposite f) `(Mult f) `(One f) `(Inverse f). *)

(*     Local Notation t := (LC.t f). *)
(*     Local Notation var := M.elt. *)

(*     Variable var_eqMixin : Equality.mixin_of var. *)
(*     Canonical var_eqType := EqType var var_eqMixin. *)

(*     Variable embedf : f → F. *)

(*     Local Notation zerolc := (@LC.zero f). *)
(*     Local Notation single := (@LC.single f). *)
(*     Local Notation add := LC.add. *)
(*     Local Notation sub := LC.sub. *)
(*     Local Notation muln := LC.muln. *)
(*     Local Notation denote := LC.denote. *)


(*     Section Spec. *)

(*       Variable embedt : t → var → F. *)
(*       Variable free_vars : t → list var. *)

(*       Record spec : Prop := *)
(*         { field_spec : FieldSpec.spec F embedf *)
(*           ; embed_zero : ∀ v, embedt zerolc v = 0%R *)
(*           ; embed_single1 : ∀ v c, embedt (single v c) v = embedf c *)
(*           ; embed_single2 : ∀ v v' c, v <> v' → embedt (single v c) v' = 0%R *)
(*           ; embed_add : ∀ v lc₁ lc₂, embedt (add lc₁ lc₂) v = ((embedt lc₁ v) + (embedt lc₂ v))%R *)
(*           ; embed_sub : ∀ v lc₁ lc₂, embedt (sub lc₁ lc₂) v = ((embedt lc₁ v) - (embedt lc₂ v))%R *)
(*           ; embed_muln : ∀ v c lc, embedt (muln c lc) v = ((embedf c) * (embedt lc v))%R *)
(*           ; embed_denote : ∀ ρ lc, embedf (denote ρ lc) = denotep (fun x => embedf (ρ x)) (fun x => embedt lc x) (free_vars lc) *)
(*           ; free_vars_uniq : ∀ lc, uniq (free_vars lc) *)
(*           ; free_vars_notin : ∀ v lc, v \notin free_vars lc → embedt lc v = 0%R *)
(*         }. *)

(*       Ltac embed_tac_lc Hlc := *)
(*         repeat ( *)
(*             rewrite (embed_zero Hlc) || *)
(*             rewrite (embed_single1 Hlc) || *)
(*             rewrite (embed_single2 Hlc) || *)
(*             rewrite (embed_add Hlc) || *)
(*             rewrite (embed_sub Hlc) || *)
(*             rewrite (embed_muln Hlc) || *)
(*             rewrite (embed_denote Hlc) *)
(*           ). *)

(*       Ltac embed_tac := *)
(*         match goal with *)
(*           | Hlc : spec |- _ => *)
(*           repeat ( *)
(*               embed_tac_field (field_spec Hlc) || *)
(*               embed_tac_lc Hlc *)
(*             ) *)

(*           | Hl : @FieldSpec.spec ?f ?feq ?p ?z ?o ?m ?o' ?i ?F ?embedf |- _ => *)
(*           repeat ( *)
(*               embed_tac_field (field_spec Hl) *)
(*             ) *)

(*           | _ => idtac *)
(*         end. *)

(*         Context (S : spec). *)

(*         Lemma zero_correct' ρ : denote ρ zerolc === zero. *)
(*         Proof. *)
(*           embed_tac. *)
(*           rewrite/denotep. apply big_rec; auto. *)
(*           move => v x _ ->. embed_tac. by rewrite mul0r add0r. *)
(*         Qed. *)

(*         Lemma single_correct' ρ x n : *)
(*           denote ρ (single x n) === mult n (ρ x). *)
(*         Proof. *)
(*           embed_tac. *)
(*           rewrite -(big_seq1 (GRing.add_monoid F) x (fun v => (embedf n * embedf (ρ v))%R)) /= -[(\sum_(j <- [:: x]) embedf n * embedf (ρ j))%R]/(denotep (λ x0 : var, embedf (ρ x0)) (fun _ => embedf n) [:: x]). *)
(*           rewrite (denotep_union₂ _ _ [:: x] (free_vars (single x n))). *)
(*           rewrite (denotep_union₁ [:: x] _ _ _ (eq_refl true) (free_vars_uniq S (single x n))). *)
(*           apply eq_bigr => y _. assert (H100: restr [eta embedt (single x n)] (free_vars (single x n)) y = restr (λ _ : var, embedf n) [:: x] y); last by rewrite H100. *)
(*           rewrite/restr in_cons in_nil orb_false_r. case e₁: (y \in free_vars (single x n)); case e₂: (y == x) => //=. *)
(*           - move/eqP: e₂ => ->. by embed_tac. *)
(*           - rewrite eq_sym in e₂. move/eqP: e₂. move => H100. by embed_tac. *)
(*           - rewrite -(embed_single1 S y n). move/eqP: e₂ => {1}->. symmetry. apply (free_vars_notin S). by apply negbT. *)
(*         Qed. *)

(*         Lemma add_correct' ρ a b : *)
(*           denote ρ (add a b) === plus (denote ρ a) (denote ρ b). *)
(*         Proof. *)
(*           embed_tac. *)
(*           rewrite (denotep_union₂ _ _ (union (free_vars a) (free_vars b)) (free_vars (add a b))). *)
(*           rewrite (denotep_union₁ _ _ _ _ (free_vars_uniq S a) (free_vars_uniq S b)). rewrite (denotep_union₁ _ _ _ _ (union_uniq _ _ (free_vars_uniq S b)) (free_vars_uniq S (add a b))). *)
(*           rewrite (denotep_union₂ _ _ (free_vars a) (free_vars b)). rewrite (denotep_union₁ _ _ _ _ (union_uniq _ _ (free_vars_uniq S b)) (free_vars_uniq S (add a b))). *)
(*           rewrite -big_split /=. apply eq_bigr => x _. rewrite restr_restr_union₁ restr_restr_union₂ -mulrDl. assert (H100: restr [eta embedt (add a b)] (free_vars (add a b)) x = (restr [eta embedt a] (free_vars a) x + restr [eta embedt b] (free_vars b) x)%R); last by rewrite H100. *)
(*           rewrite/restr. case e₁: (x \in free_vars (add a b)); case e₂: (x \in free_vars a); case e₃: (x \in free_vars b) => //=. *)
(*           - by embed_tac. *)
(*           - embed_tac. apply negbT in e₃. by rewrite (free_vars_notin S _ _ e₃). *)
(*           - embed_tac. apply negbT in e₂. by rewrite (free_vars_notin S _ _ e₂). *)
(*           - embed_tac. apply negbT in e₂. apply negbT in e₃. by rewrite (free_vars_notin S _ _ e₂) (free_vars_notin S _ _ e₃). *)
(*           - apply negbT in e₁. rewrite -(embed_add S). by rewrite (free_vars_notin S _ _ e₁). *)
(*           - apply negbT in e₁. apply negbT in e₃. rewrite -{2}(free_vars_notin S _ _ e₃). rewrite -(embed_add S). by rewrite (free_vars_notin S _ _ e₁). *)
(*           - apply negbT in e₁. apply negbT in e₂. rewrite -{2}(free_vars_notin S _ _ e₂). rewrite -(embed_add S). by rewrite (free_vars_notin S _ _ e₁). *)
(*           - by rewrite add0r. *)
(*         Qed. *)

(*         Lemma sub_correct' ρ a b : *)
(*           denote ρ (sub a b) === plus (denote ρ a) (opposite (denote ρ b)). *)
(*         Proof. *)
(*           embed_tac. *)
(*           rewrite (denotep_union₂ _ _ (union (free_vars a) (free_vars b)) (free_vars (sub a b))). *)
(*           rewrite (denotep_union₁ _ _ _ _ (free_vars_uniq S a) (free_vars_uniq S b)). rewrite (denotep_union₁ _ _ _ _ (union_uniq _ _ (free_vars_uniq S b)) (free_vars_uniq S (sub a b))). *)
(*           rewrite (denotep_union₂ _ _ (free_vars a) (free_vars b)). rewrite (denotep_union₁ _ _ _ _ (union_uniq _ _ (free_vars_uniq S b)) (free_vars_uniq S (sub a b))). *)
(*           rewrite -sumrB /=. apply eq_bigr => x _. rewrite restr_restr_union₁ restr_restr_union₂ -mulrBl. assert (H100: restr [eta embedt (sub a b)] (free_vars (sub a b)) x = (restr [eta embedt a] (free_vars a) x - restr [eta embedt b] (free_vars b) x)%R); last by rewrite H100. *)
(*           rewrite/restr. case e₁: (x \in free_vars (sub a b)); case e₂: (x \in free_vars a); case e₃: (x \in free_vars b) => //=. *)
(*           - by embed_tac. *)
(*           - embed_tac. apply negbT in e₃. by rewrite (free_vars_notin S _ _ e₃). *)
(*           - embed_tac. apply negbT in e₂. by rewrite (free_vars_notin S _ _ e₂). *)
(*           - embed_tac. apply negbT in e₂. apply negbT in e₃. by rewrite (free_vars_notin S _ _ e₂) (free_vars_notin S _ _ e₃). *)
(*           - apply negbT in e₁. rewrite -(embed_sub S). by rewrite (free_vars_notin S _ _ e₁). *)
(*           - apply negbT in e₁. apply negbT in e₃. rewrite -{2}(free_vars_notin S _ _ e₃). rewrite -(embed_sub S). by rewrite (free_vars_notin S _ _ e₁). *)
(*           - apply negbT in e₁. apply negbT in e₂. rewrite -{2}(free_vars_notin S _ _ e₂). rewrite -(embed_sub S). by rewrite (free_vars_notin S _ _ e₁). *)
(*           - by rewrite subr0. *)
(*         Qed. *)

(*         Lemma muln_correct' ρ a n : *)
(*           denote ρ (muln n a) === mult n (denote ρ a). *)
(*         Proof. *)
(*           embed_tac. *)
(*           rewrite (denotep_union₂ _ _ (free_vars a) (free_vars (muln n a))). *)
(*           rewrite (denotep_union₁ _ _ _ _ (free_vars_uniq S a) (free_vars_uniq S (muln n a))). *)
(*           rewrite mulr_sumr. apply eq_bigr => x _. rewrite mulrA. assert (H100: restr [eta embedt (muln n a)] (free_vars (muln n a)) x = (embedf n * (restr [eta embedt a] (free_vars a) x))%R); last by rewrite H100. *)
(*           rewrite/restr. case e₁: (x \in free_vars (muln n a)); case e₂: (x \in free_vars a) => //=. *)
(*           - by embed_tac. *)
(*           - embed_tac. apply negbT in e₂. by rewrite (free_vars_notin S _ _ e₂). *)
(*           - rewrite -(embed_muln S). apply negbT in e₁. by rewrite (free_vars_notin S _ _ e₁). *)
(*           - apply negbT in e₁. apply negbT in e₂. by rewrite -{2}(free_vars_notin S _ _ e₂) -(embed_muln S) (free_vars_notin S _ _ e₁). *)
(*         Qed. *)

(*       End Spec. *)


(*     Definition free_vars (a: t) : list var := *)
(*       List.rev (List.map fst (M.elements a)). *)

(*     Hypothesis (FS : FieldSpec.spec F embedf). *)

(*     Definition embedt (lc:t) (v:var) : F := *)
(*       match M.get v lc with *)
(*         | Some r => embedf r *)
(*         | None => 0%R *)
(*       end. *)

(*     Lemma in_In (A : eqType) x (l : seq A) : *)
(*       x \in l ↔ List.In x l. *)
(*     Proof. *)
(*       elim: l => [|y ys IHys] //=. *)
(*       - rewrite in_cons. split. *)
(*         * move/orP. rewrite IHys. by move=> [/eqP ->|H101]; auto. *)
(*         * rewrite -IHys; move => [->|->]. *)
(*           + by rewrite eqxx. *)
(*           + by rewrite orb_true_r. *)
(*     Qed. *)

(*     Lemma In_fst (A B : Type) (a : A) (b : B) l : *)
(*       List.In (a,b) l → List.In a (List.map fst l). *)
(*     Proof. *)
(*       elim: l => [|x xs IHxs] //=. *)
(*       - by move => [->|H101]; auto. *)
(*     Qed. *)

(*     Lemma list_norepet_uniq (A : eqType) (l : seq A) *)
(*       (H101 : Coqlib.list_norepet l) : uniq l. *)
(*     Proof. *)
(*       elim: H101 => [|x xs Hx Hxs] //= IHxs. *)
(*       - apply/andP; split; auto. *)
(*         apply/negP => H101. apply Hx. by apply in_In. *)
(*     Qed. *)

(*     Lemma list_norepet_rev (A : eqType) (l : seq A) *)
(*       (H101: Coqlib.list_norepet l) : Coqlib.list_norepet (List.rev l). *)
(*     Proof. *)
(*       elim: H101 => [|x xs Hx Hxs IHxs] //=. *)
(*       - by constructor. *)
(*       - rewrite Coqlib.list_norepet_app; repeat split; auto. *)
(*         * by repeat constructor. *)
(*         * move => y z /=. by rewrite -List.In_rev => Hy [<-|//=] H101; subst y. *)
(*     Qed. *)

(*     Lemma fold_right_foldr A B (g : B → A → A) (a : A) (l : seq B) : *)
(*       List.fold_right g a l = foldr g a l. *)
(*     Proof. by elim: l. Qed. *)

(*     Lemma foldr_map A B C (h : B → A → A) (g : C → B) (a : A) (l : seq C) : *)
(*       foldr h a (List.map g l) = foldr (fun b a => h (g b) a) a l. *)
(*     Proof. by elim: l => [|x xs /= ->]. Qed. *)

(*     Lemma foldr2 A₁ A₂ B (P : A₁ → A₂ → Prop) *)
(*           (h : B → A₁ → A₁) (g : B → A₂ → A₂) *)
(*           (a₁ : A₁) (a₂ : A₂) (l : seq B) (H101: P a₁ a₂) : *)
(*       (∀ x xs ys, (List.In x l) → P xs ys -> P (h x xs) (g x ys)) → *)
(*     P (foldr h a₁ l) (foldr g a₂ l). *)
(*     Proof. by elim: l => [|x xs IHxs] /=; auto. Qed. *)

(*     Lemma linearcomb_correct : spec embedt free_vars. *)
(*     Proof. *)
(*       split => //=. *)
(*       - rewrite /embedt /LC.zero => v. by rewrite M.gempty. *)
(*       - rewrite /embedt /LC.single => v c. by rewrite M.gss. *)
(*       - rewrite /embedt /LC.single /LC.zero => v v' c H101. rewrite M.gso; auto. by rewrite M.gempty. *)
(*       - rewrite /embedt /LC.add => v lc₁ lc₂. rewrite M.gcombine /=. *)
(*         * case: (M.get v lc₁) => [r|]; case: (M.get v lc₂) => [s|] /=; case: (equiv_dec _ _) => [|]; by embed_tac_field FS. *)
(*         * case: (equiv_dec _ _) => [|e1]; first done. *)
(*           elim e1. by embed_tac_field FS; rewrite add0r. *)
(*       - rewrite /embedt /LC.sub => v lc₁ lc₂. rewrite M.gcombine /=. *)
(*         * case: (M.get v lc₁) => [r|]; case: (M.get v lc₂) => [s|] /=; case: (equiv_dec _ _) => [|]; by embed_tac_field FS. *)
(*         * case: (equiv_dec _ _) => [|e1]; first done. *)
(*           elim e1. by embed_tac_field FS; rewrite subr0. *)
(*       - rewrite /embedt /LC.muln => v c lc. rewrite MM.gmap /=. *)
(*         * case: (M.get v lc) => [r|] /=. *)
(*           + case: (equiv_dec _ _) => [|]; by embed_tac_field FS. *)
(*           + by rewrite mulr0. *)
(*       - rewrite /embedt /LC.denote /free_vars /denotep => ρ lc. rewrite M.fold_spec -List.fold_left_rev_right fold_right_foldr BigOp.bigopE /reducebig -List.map_rev foldr_map. *)
(*         pose (P := fun x y => embedf x = y). change (P (foldr (λ y x, plus (mult y.2 (ρ y.1)) x) zero (List.rev (M.elements lc))) (foldr (λ b a, (applybig \o (λ v, BigBody v +%R true (match M.get v lc with | Some r => embedf r | None => 0 end * embedf (ρ v))%R)) b.1 a) 0%R (List.rev (M.elements lc)))). apply foldr2; rewrite /P //=. *)
(*         * by embed_tac_field FS. *)
(*         * move => [x₁ x₂] xs ys /=. rewrite -List.In_rev => H101 H102. rewrite (M.elements_complete _ _ _ H101). embed_tac_field FS. by rewrite H102. *)
(*       - move => lc. apply list_norepet_uniq. rewrite/free_vars. apply list_norepet_rev. apply M.elements_keys_norepet. *)
(*       - move => v lc H101. rewrite/embedt. case e₁: (M.get v lc) => [r|//=]. apply M.elements_correct in e₁. apply In_fst in e₁. move: e₁. rewrite List.In_rev. rewrite -(in_In var_eqType) => H102. by rewrite H102 in H101. *)
(*     Qed. *)

(*     Definition zero_correct : ∀ ρ, *)
(*       denote ρ zerolc === zero *)
(*       := zero_correct' linearcomb_correct. *)

(*     Definition single_correct : ∀ ρ x n, *)
(*       denote ρ (single x n) === mult n (ρ x) *)
(*       := single_correct' linearcomb_correct. *)

(*     Definition add_correct : ∀ ρ a b, *)
(*       denote ρ (add a b) === plus (denote ρ a) (denote ρ b) *)
(*       := add_correct' linearcomb_correct. *)

(*     Definition sub_correct : ∀ ρ a b, *)
(*       denote ρ (sub a b) === plus (denote ρ a) (opposite (denote ρ b)) *)
(*       := sub_correct' linearcomb_correct. *)

(*     Definition muln_correct : ∀ ρ a n, *)
(*       denote ρ (muln n a) === mult n (denote ρ a) *)
(*       := muln_correct' linearcomb_correct. *)

(*   End LCSpec. *)


(*   Ltac embed_tac_lc Hlc := *)
(*     repeat ( *)
(*         rewrite (embed_zero Hlc) || *)
(*         rewrite (embed_single1 Hlc) || *)
(*         rewrite (embed_single2 Hlc) || *)
(*         rewrite (embed_add Hlc) || *)
(*         rewrite (embed_sub Hlc) || *)
(*         rewrite (embed_muln Hlc) || *)
(*         rewrite (embed_denote Hlc) *)
(*       ). *)

(* End LCSpec. *)
