Require Import Utf8.
Require Import MoreLib.
Require Import Maps.

Set Implicit Arguments.

Module Make (M: TREE).

Section Make.

Context (V : Type).

  Lemma foldspec : ∀ A f (P: M.t V → A → A → Prop),
                       (∀ m a b x v, P (M.remove x m) a b → M.get x m = Some v → P m a (f b x v)) →
                       ∀ a, (∀ m, (∀ x, M.get x m = None) → P m a a) → ∀ m, P m a (M.fold f m a).
  Proof.
    intros A f P H a HE m. simpl.
    rewrite M.fold_spec.
    generalize (@M.elements_correct V m).
    generalize (@M.elements_complete V m).
    generalize (@M.elements_keys_norepet V m).
    generalize (M.elements m). intros l.
    revert m.
    induction l as [ | (x, y) l IH] using List.rev_ind.
    - simpl. intros m _ _ K. apply HE. intros x. generalize (K x). destruct (M.get x m). exact (λ K, False_ind _ (K _ refl_equal)). easy.
    - rewrite List.fold_left_app. simpl.
      intros m N R L.
      apply H.
      + rewrite List.map_app in N. simpl in N.
        apply Coqlib.list_norepet_app in N. destruct N as (N & _ & D').
        assert (¬ List.In x (List.map fst l)) as D.
        { intros K. refine (D' _ x K _ refl_equal). now left. }
        clear D'.
        apply IH. assumption.
        intros i v H0. rewrite M.grspec. destruct (M.elt_eq _ _). subst. elim D.
        rewrite List.in_map_iff. exists (x, v); split; easy.
        apply R. rewrite Coqlib.in_app. now left.
        intros i v. rewrite M.grspec. destruct (M.elt_eq _ _). exact (λ K, some_neq (eq_sym K) _).
        intros K. specialize (L _ _ K). apply Coqlib.in_app in L. destruct L as [ L | [ L | () ] ]. exact L. apply pair_eq in L. destruct L. congruence.
      + now apply R; rewrite Coqlib.in_app; right; left.
  Qed.

Section FOLD_SINGLETON.
  Let K := M.elt.
  Context {A: Type} {f: A → K → V → A}.

  Lemma fold_singleton {k v a}:
    M.fold f (M.set k v (M.empty V)) a = f a k v.
  Proof.
    generalize (λ X Y, @foldspec _ f (λ m x y, ((∀ k, M.get k m = None) ∧ x = y ∧ x = a) ∨ (∃ k' v', M.get k' m = Some v' ∧ ∀ k'' v'', M.get k'' m = Some v'' → k' = k'' ∧ v' = v'' ∧ y = f a k' v') ∨ (∃ k k', k ≠ k' ∧ M.get k m ≠ None ∧ M.get k' m ≠ None)) X a Y (M.set k v (M.empty V))).
    intros IH. destruct IH as [ IH | [ IH | IH ] ].
    - intros m x y k' v' [ (X & ? & ?) | [ (k'' & v'' & X & Y) | (k'' & k''' & X & Y & Z) ] ] G; repeat subst.
      + right. left. exists k', v'. split. exact G.
        intros k'' v'' H0.
        specialize (X k''). rewrite M.grspec in X.
        destruct (M.elt_eq k'' k'); repeat (split || congruence).
      + right. right. exists k', k''.
        rewrite M.grspec in X. destruct (M.elt_eq _ _). exact (some_neq (eq_sym X) _).
        repeat (split || congruence).
      + right. right. rewrite M.grspec in Y. rewrite M.grspec in Z.
        destruct (M.elt_eq k'' k'). exact (False_ind _ (Y eq_refl)).
        destruct (M.elt_eq k''' k'). exact (False_ind _ (Z eq_refl)).
        repeat (econstructor; eauto).
    - intros m H0. left; auto.
    - destruct IH as [IH _]. specialize (IH k). rewrite M.gsspec in IH.
      rewrite if_dec_true in IH. exact (some_neq IH _). exact eq_refl.
    - destruct IH as (k' & v' & U & U').
      specialize (U' k v).
      rewrite M.gsspec in U'. rewrite if_dec_true in U'.
      specialize (U' eq_refl).
      intuition congruence. reflexivity.
    - exfalso. destruct IH as (q & q' & Q & Hq & Hq').
      rewrite M.gsspec in Hq. rewrite M.gsspec in Hq'.
      destruct (M.elt_eq q k). destruct (M.elt_eq q' k). congruence.
      rewrite M.gempty in Hq'. auto.
      rewrite M.gempty in Hq. auto.
  Qed.

End FOLD_SINGLETON.

Section MAP.
  Let K := M.elt.
  Context {W: Type} (f: V → option W).

  Definition oset (k: K) (v: option W) (m: M.t W) : M.t W :=
    match v with Some a => M.set k a m | None => M.remove k m end.

  Lemma gosetspec k k' v m : M.get k (oset k' v m) = if M.elt_eq k k' then v else M.get k m.
  Proof. unfold oset. destruct v. apply M.gsspec. apply M.grspec. Qed.

  Definition map (m: M.t V) : M.t W := M.fold (λ a k v, oset k (f v) a) m (M.empty W).

  Lemma gmap m k : M.get k (map m) = match M.get k m with None => None | Some v => f v end.
  Proof.
    unfold map.
    apply foldspec.
    intros m0 a b x v H H0. rewrite gosetspec. rewrite M.grspec in H.
    destruct (M.elt_eq k x) eqn: EQ. subst. now rewrite H0. exact H.
    now intros m0 H; rewrite M.gempty; rewrite (H _).
  Qed.

End MAP.

End Make.
End Make.

Module PMap := Make PTree.
