Require QAP Run AbStateImpl RTLCompiler.
Require Import modpProof AbIntModp.
Import Utf8 String.
Import Integers ZArith Monads LinearCombination LargePrime.

Section ORACLE.

Context
  (O: Type)
  (oracle_gen: AbStateImpl.oracle positive O)
  (oracle_eval: @AbInt.Arith.t unit Errors.errmsg (modp.zp large_prime) U.incr → AbStateImpl.oracle (modp.zp large_prime) O)
.

Existing Instances QAP.incr_refl QAP.incr_trans U.incr_refl U.incr_trans.
Existing Instances zpEquiv zpPlus zpZero zpOpp zpMult zpOne zpInv.

Let sa_gen (ge: RTL.genv) := AbStateImpl.sa ge (QAP.cQAP (@modp.zconst large_prime)) oracle_gen.
Let sa_eval (ge: RTL.genv) :=
  let arith := modp_arith large_prime in
  AbStateImpl.sa ge arith (oracle_eval arith).

Existing Instances AbStateImpl.incr_refl AbStateImpl.incr_trans.

Context (cse cp dc: bool) (prog: Csyntax.program) (fuel: nat).

Definition qap_gen :=
  match RTLCompiler.transf_c_program cse cp dc prog with
  | Errors.OK rtl =>
      let ge := Globalenvs.Genv.globalenv rtl in
      Compiler.time "qap_gen" (Run.run (sa_gen ge) rtl) fuel
  | Errors.Error msg => Error msg
  end.

Definition eval :=
  match RTLCompiler.transf_c_program cse cp dc prog with
  | Errors.OK rtl =>
      let ge := Globalenvs.Genv.globalenv rtl in
      Compiler.time "qap_eval" (Run.run (sa_eval ge) rtl) fuel
  | Errors.Error msg => Error msg
  end.

End ORACLE.
