Require Import Interval modp LargePrime.
Import Utf8 Bool ZArith Psatz.
Require Import PP.PpsimplZ.
Import MoreLib.
Local Open Scope Z_scope.

Set Implicit Arguments.

Import Interval.Notations.
Require Import Integers.
Local Open Scope interval_scope.

Definition project (z: zp large_prime) (i: Interval.interval) : option Z :=
  let q := Zpos large_prime in
  let z := Z.of_N z in
  let '(k, r) := Z.div_eucl (imin i - z) q in
  let res := if r =? 0 then z + k * q else z + (k + 1) * q in
  if res <=? imax i
  then Some res
  else None.

Lemma project_range :
  ∀ (x : zp large_prime) (xr : Interval.interval)
    (x' : Z), project x xr = Some x' → is_true (x' ∈ xr).
Proof.
unfold project. intros [x Hx] [xm xM] x'.
unfold Interval.denote, is_true. simpl.
apply N.ltb_lt, N2Z.inj_lt in Hx.
elim_div.
intros H. specialize (H large_prime_nz).
destruct H as [K [ H | H ] ]. 2: exfalso; lia.
assert (0 <= Z.of_N x). destruct x; discriminate.
intros H1.
repeat match goal with | H : appcontext[ match ?x with _ => _ end ] |- _ => destruct x eqn: ?; try (refine (some_neq (eq_sym _) _); eassumption); try apply some_eq in H end;
subst; ppsimpl;
intros;
repeat match goal with
| H : _ ∧ _ |- _ => destruct H
| H : _ ↔ _ |- _ => destruct H
| H : True → _ |- _ => specialize (H I)
| H : _ → True |- _ => clear H
| H : False → _ |- _ => clear H
| H : ?a ≠ 0 → _ |- _ => let Q := fresh in assert (a ≠ 0) as Q by lia; specialize (H Q); clear Q
| H : _ ∨ _ |- _ => destruct H; [ | lia ]
end; subst;
lia.
Qed.

Lemma project_eqm x xr x' :
  project x xr = Some x' →
  ∃ k, x' = Z.of_N x + k * Zpos large_prime.
Proof.
  unfold project. elim_div.
  intros H. specialize (H large_prime_nz).
  repeat
  match goal with
  | |- appcontext[ if ?x =? 0 then _ else _ ] =>
    generalize (Z.eqb_spec x 0); inversion 1
  | |- appcontext[ if ?b then _ else _ ] => destruct b eqn: ?; intros
  | H : None = Some _ |- _ => exact (some_neq (eq_sym H) _)
  | H : Some _ = Some _ |- _ => apply some_eq in H; subst
  end.
  exists z. ring.
  exists (z + 1). ring.
Qed.

Lemma denoteP {i x} : x ∈ i → imin i <= x <= imax i.
Proof.
  now unfold denote, is_true; rewrite andb_true_iff, ! Z.leb_le.
Qed.

Lemma lebP {i j} : Interval.leb i j → imin j <= imin i ∧ imax i <= imax j.
Proof.
  now unfold Interval.leb, is_true; rewrite andb_true_iff, ! Z.leb_le.
Qed.

Lemma project_subrange (x: zp large_prime) (r: interval) (m: Z) (a b: Z) :
  project x r = Some a →
  project x {| imin := 0; imax := m |} = Some b →
  Interval.leb r {| imin := 0; imax := m |} →
  0 <= m < Zpos large_prime →
  a = b.
Proof.
  intros H. destruct (project_eqm x r H) as ( k & K ). subst a.
  intros H'. destruct (project_eqm x _ H') as ( k' & K' ). subst b.
  pose proof (project_range x r H) as Hr. apply denoteP in Hr.
  pose proof (project_range x _ H') as Hr'. apply denoteP in Hr'. simpl in Hr'.
  clear H H'.
  intros LE. apply lebP in LE. simpl in LE.
  intros H.
  unfold large_prime in *.
  nia.
Qed.

Lemma project_ex (x: zp large_prime) xr :
  (∃ k, Z.of_N x + k * Zpos large_prime ∈ xr) →
  project x xr ≠ None.
Proof.
  unfold denote, size.
  destruct x as [x Hx]. destruct xr as (xm & xM). simpl.
  intros (k & Hk). unfold is_true in Hk. rewrite andb_true_iff in Hk. destruct Hk as (Hm & HM).
  rewrite Z.leb_le in Hm.
  rewrite Z.leb_le in HM.
  unfold project. simpl. elim_div.
  intros H. specialize (H large_prime_nz). destruct H as (H & [ H' | ?]). 2: lia.
  apply ltb_lt in Hx.
  assert (0 <= xM - xm). cut (0 <= xM - xm). 2: lia. destruct (xM - xm); lia.
  repeat
  match goal with
  | |- appcontext[ if ?x =? 0 then _ else _ ] => generalize (Z.eqb_spec x 0); inversion 1; subst
  | |- appcontext[ if ?x <=? ?y then _ else _ ] => generalize (Z.leb_spec0 x y); inversion 1; subst
  end; try refine (λ H, some_neq H _).
  lia.
  assert (k = z + 1). nia. subst. lia.
Qed.

Lemma same_range_diff a i j :
  i ∈ a → j ∈ a → j - i <= Z.of_N (size a).
Proof.
  unfold denote, size, is_true.
  intros H H0.
  destruct (imax a - imin a) eqn: ?; ppsimpl; lia.
Qed.

Lemma project_ex_strong (x: zp large_prime) xr :
  (size xr < Npos large_prime)%N →
  ∀ k,
  Z.of_N x + k * Zpos large_prime ∈ xr →
  project x xr = Some (Z.of_N x + k * Zpos large_prime).
Proof.
  intros H k H0. destruct (not_none (project_ex _ _ (ex_intro _ k H0))) as (z & hz). rewrite hz. f_equal.
  pose proof (project_range _ _ hz) as hr.
  destruct (project_eqm _ _ hz) as (k' & ->).
  repeat f_equal.
  apply N2Z.inj_lt in H. simpl in H.
  pose proof (same_range_diff _ _ _ H0 hr).
  pose proof (same_range_diff _ _ _ hr H0).
  nia.
Qed.

Lemma project_const x :
  project (zconst large_prime x) (Interval.const x) = Some x.
Proof.
  destruct (λ X, not_none (project_ex (zconst large_prime x) (Interval.const x) X)) as (x' & h).
  unfold zconst. destruct x as [ | x | x ].
  exists 0. exact eq_refl.
  simpl. rewrite N2Z.inj_mod. 2: lia. simpl.
  exists (Zpos x / Zpos large_prime). elim_div. intros h. specialize (h large_prime_nz). destruct h as (h & [ h' | ? ]). 2: lia.
  apply Interval.const_iff. lia.
  apply neg_elim; intros h.
  apply N2Z.inj_iff in h; simpl in h. rewrite N2Z.inj_mod in h. simpl in h. apply Z_div_exact_full_2 in h.
  exists (- (Zpos x / Zpos large_prime)). apply Interval.const_iff. lia. apply large_prime_nz. lia.
  rewrite N2Z.inj_sub. 2: lia.
  rewrite ! N2Z.inj_lt in h. simpl in h. rewrite N2Z.inj_mod in h. 2: lia. simpl in h.
  unfold const, zp_to_n. rewrite N2Z.inj_mod. 2: lia. unfold Z.of_N.
  elim_div. intros H. specialize (H large_prime_nz). destruct H as (H & [ H' | ? ]). 2: lia.
  exists (- z - 1). apply Interval.const_iff. lia.
  pose proof (Interval.const_exact _ _ (project_range _ _ h)).
  congruence.
Qed.

Lemma project_add x xr x' y yr y' :
  project x xr = Some x' →
  project y yr = Some y' →
  (Interval.size (Interval.add xr yr) < Npos large_prime)%N →
  project (add x y) (Interval.add xr yr) = Some (x' + y').
Proof.
  intros Hx Hy sz.
  destruct (project_eqm _ _ Hx) as (u & ->).
  destruct (project_eqm _ _ Hy) as (v & ->).
  assert (∃ k, Z.of_N (add x y) + k * Zpos large_prime = (Z.of_N x + u * Zpos large_prime) + (Z.of_N y + v * Zpos large_prime)) as K.
  unfold add. simpl. apply overflow_elim. intros z [ (h & <-) | h ]. exists (u + v). lia. exists ( u + v + 1 ). lia.
  destruct K as (k & K).
  pose proof (Interval.add_correct _ _ _ _ (project_range _ _ Hx) (project_range _ _ Hy)) as IN.
  rewrite <- K in IN.
  pose proof (project_ex_strong (add x y) _ sz _ IN).
  congruence.
Qed.

Lemma project_sub x xr x' y yr y' :
  project x xr = Some x' →
  project y yr = Some y' →
  (Interval.size (Interval.sub xr yr) < Npos large_prime)%N →
  project (sub x y) (Interval.sub xr yr) = Some (x' - y').
Proof.
  intros Hx Hy sz.
  destruct (project_eqm _ _ Hx) as (u & ->).
  destruct (project_eqm _ _ Hy) as (v & ->).
  assert (∃ k, Z.of_N (sub x y) + k * Zpos large_prime = (Z.of_N x + u * Zpos large_prime) - (Z.of_N y + v * Zpos large_prime)) as K.
  unfold sub. simpl. apply overflow_elim. intros z [ (h & <-) | h ].
  apply neg_elim. intros ->. simpl. exists (u - v). lia. exists (u - v - 1). lia.
  revert h. apply neg_elim. intros ->. simpl. exists (u - v + 1). lia. exists (u - v). lia.
  destruct K as (k & K).
  pose proof (Interval.sub_correct _ _ _ _ (project_range _ _ Hx) (project_range _ _ Hy)) as IN.
  rewrite <- K in IN.
  pose proof (project_ex_strong (sub x y) _ sz _ IN).
  congruence.
Qed.

Lemma project_muln x' y yr y':
   project y yr = Some y' →
   (Interval.size (Interval.muln x' yr) < N.pos large_prime)%N →
   project (mul (zconst large_prime x') y) (Interval.muln x' yr) = Some (x' * y').
Proof.
  Opaque Z.mul N.sub Z.sub.
  intros Hy sz.
  destruct (project_eqm _ _ Hy) as (v & ->).
  assert (∃ k, Z.of_N (mul (zconst large_prime x') y) + k * Zpos large_prime = x' * (Z.of_N y + v * Zpos large_prime)) as K.
  unfold mul, zconst. simpl.
  destruct x'. exists 0. exact eq_refl.
  rewrite mod_mul_spec.
  simpl. rewrite N2Z.inj_mod. 2: lia. rewrite N2Z.inj_mul. simpl.
  rewrite N2Z.inj_mod. 2: lia. simpl. rewrite Z.mul_mod_idemp_l. 2: exact large_prime_nz.
  elim_div. intros h. specialize (h large_prime_nz).
  exists (z + Zpos p * v). lia.
  rewrite mod_mul_spec.
  apply neg_elim. intros hp. clear -hp. unfold const in hp. simpl in hp. rewrite (N.mod_divides) in hp. 2: lia. destruct hp as (k & hp).
  simpl. apply N2Z.inj_iff in hp. exists (- Z.of_N k * Z.of_N y - Z.of_N k * v * Zpos large_prime). nia.
  intros hp. simpl.
  rewrite N2Z.inj_mod. 2: lia. rewrite N2Z.inj_mul, N2Z.inj_sub, N2Z.inj_mod. 2: lia.
  2: generalize (N.mod_upper_bound (Npos p) (Npos large_prime)); lia.
  simpl.
  rewrite Z.mul_sub_distr_r. clear sz hp Hy.
  assert (∀ a b c, (b * a - c) mod b = (- c) mod b) as X.
  intros a b c. replace (b * a - c) with (- c + a * b) by lia. now rewrite Z_mod_plus_full.
  rewrite X. clear X.
  elim_div. intros h. specialize (h large_prime_nz). intros h'. specialize (h' large_prime_nz).
  destruct h as (h & [ hh | ?] ). 2: lia.
  destruct h' as (h' & [ hh' | ?] ). 2: lia.
  assert (z0 = - (z2 * Z.of_N y) - Zpos large_prime * z). lia. subst.
  assert (Zneg p = - Zpos large_prime * z1 - z2) as X by lia. rewrite X. clear h X p.
  exists (z - z2 * v - Z.of_N y * z1 - Zpos large_prime * v * z1). lia.
  destruct K as (k & K).
  pose proof (Interval.muln_correct x' _ _ (project_range _ _ Hy)) as IN.
  rewrite <- K in IN.
  pose proof (project_ex_strong (mul (zconst large_prime x') y) _ sz _ IN).
  congruence.
  Transparent Z.mul N.sub Z.sub.
Qed.

Lemma project_mul x xr x' y yr y' :
  project x xr = Some x' →
  project y yr = Some y' →
  (Interval.size (Interval.mul xr yr) < Npos large_prime)%N →
  project (mul x y) (Interval.mul xr yr) = Some (x' * y').
Proof.
  intros Hx Hy sz.
  destruct (project_eqm _ _ Hx) as (u & ->).
  destruct (project_eqm _ _ Hy) as (v & ->).
  assert (∃ k, Z.of_N (mul x y) + k * Zpos large_prime = (Z.of_N x + u * Zpos large_prime) * (Z.of_N y + v * Zpos large_prime)) as K.
  unfold mul. simpl. rewrite mod_mul_spec.
  rewrite N2Z.inj_mod. 2: lia. rewrite N2Z.inj_mul. elim_div.
  intros h. specialize (h large_prime_nz).
  exists (z + u * Z.of_N y + v * Z.of_N x + u * v * Zpos large_prime). lia.
  destruct K as (k & K).
  pose proof (Interval.mul_correct _ _ _ _ (project_range _ _ Hx) (project_range _ _ Hy)) as IN.
  rewrite <- K in IN.
  pose proof (project_ex_strong (mul x y) _ sz _ IN).
  congruence.
Qed.

Lemma project_fun x xr xr' a a' :
  project x xr = Some a →
  project x xr' = Some a' →
  Interval.leb (Interval.sub xr xr') {| imin := 1 - Zpos large_prime; imax := Zpos large_prime - 1 |} →
  a = a'.
Proof.
Opaque Z.sub.
  unfold Interval.sub, Interval.leb. simpl.
  intros h h' LE.
  destruct (project_eqm _ _ h) as (u & ->).
  destruct (project_eqm _ _ h') as (v & ->).
  pose proof (project_range _ _ h).
  pose proof (project_range _ _ h').
  clear h h'. unfold Interval.denote, is_true in *.
  assert (- Zpos large_prime < (u - v) * Zpos large_prime < Zpos large_prime). ppsimpl. lia.
  cut (u - v = 0). nia.
  nia.
Transparent Z.sub.
Qed.

Lemma project_inj x xr y yr a b :
  project x xr = Some a →
  project y yr = Some b →
  x ≠ y → a ≠ b.
Proof.
  intros h h' NE EQ. apply NE, zp_eq, N2Z.inj.
  destruct (project_eqm _ _ h) as (u & ->).
  destruct (project_eqm _ _ h') as (v & ->).
  clear -EQ.
  destruct x as [x hx], y as [y hy]. simpl in *.
  apply N.ltb_lt, N2Z.inj_lt in hx. apply N.ltb_lt, N2Z.inj_lt in hy. simpl in *.
  destruct (Zdiv_mod_unique_2 (Zpos large_prime) u v (Z.of_N x) (Z.of_N y)); try red; try lia.
Qed.

(* Split a natural number after n bits: (efficient) division by 2^n. *)
Fixpoint cutN (n: nat) (x: N) : N * N :=
  match n with
  | O => (x, N0)
  | S n' =>
    match x with
    | N0 => (x, x)
    | Npos p =>
      match p with
      | 1 => (N0, x)
      | p'~0 => let '(q, r) := cutN n' (Npos p') in (q, 2 * r)%N
      | p'~1 => let '(q, r) := cutN n' (Npos p') in (q, 1 + 2 * r)%N
      end%positive
    end
  end.

Lemma cutN_correct n :
  ∀ x q r,
  cutN n x = (q, r) →
  (x = r + q * N.shiftl_nat 1 n)%N.
Proof.
  induction n as [ | n IH ]; simpl; intros x q r H.
  apply pair_eq in H. destruct H; subst. simpl. ppsimpl. lia.
  destruct x as [ | p ]. apply pair_eq in H. destruct H; subst. exact eq_refl.
  destruct p as [ p' | p' | ]; try (specialize (IH (Npos p')); destruct (cutN n _) as (q',  [ | r' ]));
  apply pair_eq in H; destruct H; subst;
  try (specialize (IH _ _ eq_refl); rewrite Ndigits.Nshiftl_nat_S, N.double_spec; lia).
  lia.
Qed.
Global Opaque cutN.

Definition split8 (a: zp large_prime) : N * N * N * N * N * N * N * N :=
  let m : nat := 32%nat in
  let q : N := a in
  let (q, b₀) := cutN m q in
  let (q, b₁) := cutN m q in
  let (q, b₂) := cutN m q in
  let (q, b₃) := cutN m q in
  let (q, b₄) := cutN m q in
  let (q, b₅) := cutN m q in
  let (q, b₆) := cutN m q in
  let (q, b₇) := cutN m q in
  (b₇, b₆, b₅, b₄, b₃, b₂, b₁, b₀).

Definition merge8 (b: N * N * N * N * N * N * N * N) : N :=
  (let m : N := N.shiftl 1 32 in
  let '(b₇, b₆, b₅, b₄, b₃, b₂, b₁, b₀) := b in
  let n := 0 in
  let n := n * m + b₇ in
  let n := n * m + b₆ in
  let n := n * m + b₅ in
  let n := n * m + b₄ in
  let n := n * m + b₃ in
  let n := n * m + b₂ in
  let n := n * m + b₁ in
  let n := n * m + b₀ in
  n)%N.

Lemma merge8_split8 (a: zp large_prime) :
  merge8 (split8 a) = a.
Proof.
  unfold split8, merge8.
  destruct a as [a Ha].
  simpl zp_to_n.
  apply ltb_lt, N2Z.inj_lt in Ha.
  repeat match goal with
  | |- appcontext[ let (u,v) := cutN ?a ?b in _] =>
    let H := fresh in
    let u := fresh u in
    let v := fresh v in
    generalize (cutN_correct a b);
    destruct (cutN a b) as (u,v);
    intros H; specialize (H _ _ eq_refl); subst
  end.
  apply N2Z.inj. repeat ( rewrite N2Z.inj_add, N2Z.inj_mul in * ).
  cut (b₇ = N0). intros ->. simpl. ring.
  unfold large_prime in *. simpl in *. lia.
Qed.
