{ stdenv, fetchurl, tools, coq, menhir }:

stdenv.mkDerivation {
  pname = "compcert";
  version = "2.4";

  src = fetchurl {
    url = http://compcert.inria.fr/release/compcert-2.4.tgz;
    sha256 = "1qrb1cplx3v5wxn1c46kx67v1j52yznvjm2hkrsdybphhki2pyia";
  };

  buildInputs = [ coq coq.ocaml menhir ];

  enableParallelBuilding = true;

  configurePhase = ''
    substituteInPlace ./configure --replace '{toolprefix}gcc' '{toolprefix}cc'
    ./configure -prefix $out -toolprefix ${tools}/bin/ '' +
    (if stdenv.isDarwin then "ia32-macosx" else "ia32-linux");

  installTargets = "documentation install";

  postInstall = ''
    mkdir -p $out/share/doc/compcert
    mv doc/html $out/share/doc/compcert/
    mkdir -p $out/lib/coq/${coq.coq-version}/user-contrib/compcert/
    mv backend cfrontend common cparser driver flocq ia32 lib \
      $out/lib/coq/${coq.coq-version}/user-contrib/compcert/
  '';

  meta = {
    homepage = "http://compcert.inria.fr/";
    license = stdenv.lib.licenses.inria-compcert;
  };

}
