{ stdenv, fetchurl, unzip, ecm, gmp, coq }:

stdenv.mkDerivation {
  name = "coq-coqprime-8.4";
  src = fetchurl {
    url = https://gforge.inria.fr/frs/download.php/file/34945/coqprime_v84.zip;
    sha256 = "0bc845zlkvikfhx8b3s6xysh9xdapqhgi0jfb1b11rh4r4nwhzp7";
  };

  nativeBuildInputs = [ unzip ];
  buildInputs = [ ecm gmp coq ] ++ (with coq; [ ocaml camlp5 ]);

  configurePhase = ''
    rm -vf Makefile
    find -name '*.vo' | xargs rm -vf
    find -name '*.v.d' | xargs rm -vf
    coq_makefile -o Makefile \
      -R . coqprime \
      Tactic/Tactic.v \
      N/NatAux.v \
      List/*.v \
      Z/*.v \
      PrimalityTest/*.v \
      elliptic/SMain.v elliptic/ZEll.v elliptic/GZnZ.v \
      num/NEll.v num/W.v num/Mod_op.v num/Lucas.v num/Pock.v \
      num/Bits.v num/montgomery.v num/MEll.v \
      examples/BasePrimes.v examples/PocklingtonRefl.v
  '';

  buildFlags = "all html";

  enableParallelBuilding = true;

  postBuild = "cd gencertif; make; cd ..";

  installFlags = [
    "COQLIBINSTALL=$(out)/lib/coq/8.4/user-contrib/"
    "COQDOCINSTALL=$(out)/share/doc/coq/8.4/user-contrib/"
  ];
  installTargets = [ "install" "install-doc" ];

  preInstall = ''
    mkdir -p $out/bin
    cp gencertif/pocklington $out/bin/
  '';

  meta = {
    homepage = http://coqprime.gforge.inria.fr/;
    license = stdenv.lib.licenses.lgpl21;
  };
}
